@@ -8,7 +8,7 @@
 /// The caller must free non-const array results.
 ///
 /// \copyright
-/// Copyright (c) 2014-2016 Josh Blum
+/// Copyright (c) 2014-2017 Josh Blum
 /// Copyright (c) 2016-2016 Bastille Networks
 /// SPDX-License-Identifier: BSL-1.0
 ///
@@ -84,19 +84,17 @@ SOAPY_SDR_API SoapySDRDevice *SoapySDRDevice_make(const SoapySDRKwargs *args);
  * For every call to make, there should be a matched call to unmake.
  *
  * \param args a markup string of key/value arguments
- * \return a pointer to a new Device object
+ * \return a pointer to a new Device object or null for error
  */
 SOAPY_SDR_API SoapySDRDevice *SoapySDRDevice_makeStrArgs(const char *args);
 
 /*!
  * Unmake or release a device object handle.
  *
- * \note This call is not thread safe. Implementations calling into unmake
- * from multiple threads should protect this call with a mutex.
- *
  * \param device a pointer to a device object
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_unmake(SoapySDRDevice *device);
+SOAPY_SDR_API int SoapySDRDevice_unmake(SoapySDRDevice *device);
 
 /*******************************************************************
  * Identification API
@@ -186,7 +184,8 @@ SOAPY_SDR_API bool SoapySDRDevice_getFullDuplex(const SoapySDRDevice *device, co
  * \param direction the channel direction RX or TX
  * \param channel an available channel on the device
  * \param [out] length the number of format strings
- * \return a list of allowed format strings
+ * \return a list of allowed format strings.
+ *  See SoapySDRDevice_setupStream() for the format syntax.
  */
 SOAPY_SDR_API char **SoapySDRDevice_getStreamFormats(const SoapySDRDevice *device, const int direction, const size_t channel, size_t *length);
 
@@ -214,34 +213,55 @@ SOAPY_SDR_API SoapySDRArgInfo *SoapySDRDevice_getStreamArgsInfo(const SoapySDRDe
 
 /*!
  * Initialize a stream given a list of channels and stream arguments.
+ * The implementation may change switches or power-up components.
  * All stream API calls should be usable with the new stream object
  * after setupStream() is complete, regardless of the activity state.
  *
- * Format string markup guidelines:
- *  - C means complex
- *  - F means floating point
- *  - S means signed integer
- *  - U means unsigned integer
- *  - number float/int size in bytes (complex is 2x this size)
+ * The API allows any number of simultaneous TX and RX streams, but many dual-channel
+ * devices are limited to one stream in each direction, using either one or both channels.
+ * This call will return an error if an unsupported combination is requested,
+ * or if a requested channel in this direction is already in use by another stream.
  *
- * Example format strings:
- *  - CF32 complex float32 (8 bytes per element)
- *  - CS16 complex int16 (4 bytes per element)
- *  - CS12 complex int12 (3 bytes per element)
- *  - CS4 complex int4 (1 byte per element)
- *  - S32 int32 (4 bytes per element)
- *  - U8 uint8 (1 byte per element)
- *
- * Recommended keys to use in the args dictionary:
- *  - "WIRE" - format of the samples between device and host
+ * When multiple channels are added to a stream, they are typically expected to have
+ * the same sample rate. See SoapySDRDevice_setSampleRate().
  *
  * \param device a pointer to a device instance
- * \param [out] stream the opaque pointer to a stream handle
- * \param direction the channel direction RX or TX
- * \param format the desired buffer format in read/writeStream()
- * \param channels a list of channels for empty for automatic
+ * \param [out] stream the opaque pointer to a stream handle.
+ * \parblock
+ *
+ * The returned stream is not required to have internal locking, and may not be used
+ * concurrently from multiple threads.
+ * \endparblock
+ *
+ * \param direction the channel direction (`SOAPY_SDR_RX` or `SOAPY_SDR_TX`)
+ * \param format A string representing the desired buffer format in read/writeStream()
+ * \parblock
+ *
+ * The first character selects the number type:
+ *   - "C" means complex
+ *   - "F" means floating point
+ *   - "S" means signed integer
+ *   - "U" means unsigned integer
+ *
+ * The type character is followed by the number of bits per number (complex is 2x this size per sample)
+ *
+ *  Example format strings:
+ *   - "CF32" -  complex float32 (8 bytes per element)
+ *   - "CS16" -  complex int16 (4 bytes per element)
+ *   - "CS12" -  complex int12 (3 bytes per element)
+ *   - "CS4" -  complex int4 (1 byte per element)
+ *   - "S32" -  int32 (4 bytes per element)
+ *   - "U8" -  uint8 (1 byte per element)
+ *
+ * \endparblock
+ * \param channels a list of channels or empty for automatic
  * \param numChans the number of elements in the channels array
  * \param args stream args or empty for defaults
+ * \parblock
+ *
+ *   Recommended keys to use in the args dictionary:
+ *    - "WIRE" - format of the samples between device and host
+ * \endparblock
  * \return 0 for success or error code on failure
  */
 SOAPY_SDR_API int SoapySDRDevice_setupStream(SoapySDRDevice *device,
@@ -256,8 +276,9 @@ SOAPY_SDR_API int SoapySDRDevice_setupStream(SoapySDRDevice *device,
  * Close an open stream created by setupStream
  * \param device a pointer to a device instance
  * \param stream the opaque pointer to a stream handle
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_closeStream(SoapySDRDevice *device, SoapySDRStream *stream);
+SOAPY_SDR_API int SoapySDRDevice_closeStream(SoapySDRDevice *device, SoapySDRStream *stream);
 
 /*!
  * Get the stream's maximum transmission unit (MTU) in number of elements.
@@ -612,8 +633,9 @@ SOAPY_SDR_API int SoapySDRDevice_setDCOffset(SoapySDRDevice *device, const int d
  * \param channel an available channel on the device
  * \param [out] offsetI the relative correction (1.0 max)
  * \param [out] offsetQ the relative correction (1.0 max)
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_getDCOffset(const SoapySDRDevice *device, const int direction, const size_t channel, double *offsetI, double *offsetQ);
+SOAPY_SDR_API int SoapySDRDevice_getDCOffset(const SoapySDRDevice *device, const int direction, const size_t channel, double *offsetI, double *offsetQ);
 
 /*!
  * Does the device support frontend IQ balance correction?
@@ -642,8 +664,37 @@ SOAPY_SDR_API int SoapySDRDevice_setIQBalance(SoapySDRDevice *device, const int
  * \param channel an available channel on the device
  * \param [out] balanceI the relative correction (1.0 max)
  * \param [out] balanceQ the relative correction (1.0 max)
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_getIQBalance(const SoapySDRDevice *device, const int direction, const size_t channel, double *balanceI, double *balanceQ);
+SOAPY_SDR_API int SoapySDRDevice_getIQBalance(const SoapySDRDevice *device, const int direction, const size_t channel, double *balanceI, double *balanceQ);
+
+/*!
+ * Does the device support frontend frequency correction?
+ * \param device a pointer to a device instance
+ * \param direction the channel direction RX or TX
+ * \param channel an available channel on the device
+ * \return true if frequency corrections are supported
+ */
+SOAPY_SDR_API bool SoapySDRDevice_hasFrequencyCorrection(const SoapySDRDevice *device, const int direction, const size_t channel);
+
+/*!
+ * Fine tune the frontend frequency correction.
+ * \param device a pointer to a device instance
+ * \param direction the channel direction RX or TX
+ * \param channel an available channel on the device
+ * \param value the correction in PPM
+ * \return an error code or 0 for success
+ */
+SOAPY_SDR_API int SoapySDRDevice_setFrequencyCorrection(SoapySDRDevice *device, const int direction, const size_t channel, const double value);
+
+/*!
+ * Get the frontend frequency correction value.
+ * \param device a pointer to a device instance
+ * \param direction the channel direction RX or TX
+ * \param channel an available channel on the device
+ * \return the correction value in PPM
+ */
+SOAPY_SDR_API double SoapySDRDevice_getFrequencyCorrection(const SoapySDRDevice *device, const int direction, const size_t channel);
 
 /*******************************************************************
  * Gain API
@@ -891,6 +942,7 @@ SOAPY_SDR_API double SoapySDRDevice_getSampleRate(const SoapySDRDevice *device,
 
 /*!
  * Get the range of possible baseband sample rates.
+ * \deprecated replaced by getSampleRateRange()
  * \param device a pointer to a device instance
  * \param direction the channel direction RX or TX
  * \param channel an available channel on the device
@@ -899,6 +951,16 @@ SOAPY_SDR_API double SoapySDRDevice_getSampleRate(const SoapySDRDevice *device,
  */
 SOAPY_SDR_API double *SoapySDRDevice_listSampleRates(const SoapySDRDevice *device, const int direction, const size_t channel, size_t *length);
 
+/*!
+ * Get the range of possible baseband sample rates.
+ * \param device a pointer to a device instance
+ * \param direction the channel direction RX or TX
+ * \param channel an available channel on the device
+ * \param [out] length the number of sample rates
+ * \return a list of sample rate ranges in samples per second
+ */
+SOAPY_SDR_API SoapySDRRange *SoapySDRDevice_getSampleRateRange(const SoapySDRDevice *device, const int direction, const size_t channel, size_t *length);
+
 /*******************************************************************
  * Bandwidth API
  ******************************************************************/
@@ -1043,8 +1105,9 @@ SOAPY_SDR_API long long SoapySDRDevice_getHardwareTime(const SoapySDRDevice *dev
  * \param device a pointer to a device instance
  * \param timeNs time in nanoseconds
  * \param what optional argument
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_setHardwareTime(SoapySDRDevice *device, const long long timeNs, const char *what);
+SOAPY_SDR_API int SoapySDRDevice_setHardwareTime(SoapySDRDevice *device, const long long timeNs, const char *what);
 
 /*!
  * Set the time of subsequent configuration calls.
@@ -1054,8 +1117,9 @@ SOAPY_SDR_API void SoapySDRDevice_setHardwareTime(SoapySDRDevice *device, const
  * \param device a pointer to a device instance
  * \param timeNs time in nanoseconds
  * \param what optional argument
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_setCommandTime(SoapySDRDevice *device, const long long timeNs, const char *what);
+SOAPY_SDR_API int SoapySDRDevice_setCommandTime(SoapySDRDevice *device, const long long timeNs, const char *what);
 
 /*******************************************************************
  * Sensor API
@@ -1143,8 +1207,9 @@ SOAPY_SDR_API char **SoapySDRDevice_listRegisterInterfaces(const SoapySDRDevice
  * \param name the name of a available register interface
  * \param addr the register address
  * \param value the register value
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeNamedRegister(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned value);
+SOAPY_SDR_API int SoapySDRDevice_writeRegister(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned value);
 
 /*!
  * Read a register on the device given the interface name.
@@ -1153,27 +1218,32 @@ SOAPY_SDR_API void SoapySDRDevice_writeNamedRegister(SoapySDRDevice *device, con
  * \param addr the register address
  * \return the register value
  */
-SOAPY_SDR_API unsigned SoapySDRDevice_readNamedRegister(const SoapySDRDevice *device, const char *name, const unsigned addr);
+SOAPY_SDR_API unsigned SoapySDRDevice_readRegister(const SoapySDRDevice *device, const char *name, const unsigned addr);
 
 /*!
- * Write a register on the device.
- * This can represent a register on a soft CPU, FPGA, IC;
+ * Write a memory block on the device given the interface name.
+ * This can represent a memory block on a soft CPU, FPGA, IC;
  * the interpretation is up the implementation to decide.
- * \deprecated replaced by writeRegister(name)
  * \param device a pointer to a device instance
- * \param addr the register address
- * \param value the register value
+ * \param name the name of a available memory block interface
+ * \param addr the memory block start address
+ * \param value the memory block content
+ * \param length the number of words in the block
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeRegister(SoapySDRDevice *device, const unsigned addr, const unsigned value);
+SOAPY_SDR_API int SoapySDRDevice_writeRegisters(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned *value, const size_t length);
 
 /*!
- * Read a register on the device.
- * \deprecated replaced by readRegister(name)
+ * Read a memory block on the device given the interface name.
+ * Pass the number of words to be read in via length;
+ * length will be set to the number of actual words read.
  * \param device a pointer to a device instance
- * \param addr the register address
- * \return the register value
+ * \param name the name of a available memory block interface
+ * \param addr the memory block start address
+ * \param [inout] length number of words to be read from memory block
+ * \return the memory block content
  */
-SOAPY_SDR_API unsigned SoapySDRDevice_readRegister(const SoapySDRDevice *device, const unsigned addr);
+SOAPY_SDR_API unsigned *SoapySDRDevice_readRegisters(const SoapySDRDevice *device, const char *name, const unsigned addr, size_t *length);
 
 /*******************************************************************
  * Settings API
@@ -1193,8 +1263,9 @@ SOAPY_SDR_API SoapySDRArgInfo *SoapySDRDevice_getSettingInfo(const SoapySDRDevic
  * \param device a pointer to a device instance
  * \param key the setting identifier
  * \param value the setting value
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeSetting(SoapySDRDevice *device, const char *key, const char *value);
+SOAPY_SDR_API int SoapySDRDevice_writeSetting(SoapySDRDevice *device, const char *key, const char *value);
 
 /*!
  * Read an arbitrary setting on the device.
@@ -1222,8 +1293,9 @@ SOAPY_SDR_API SoapySDRArgInfo *SoapySDRDevice_getChannelSettingInfo(const SoapyS
  * \param channel an available channel on the device
  * \param key the setting identifier
  * \param value the setting value
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeChannelSetting(SoapySDRDevice *device, const int direction, const size_t channel, const char *key, const char *value);
+SOAPY_SDR_API int SoapySDRDevice_writeChannelSetting(SoapySDRDevice *device, const int direction, const size_t channel, const char *key, const char *value);
 
 /*!
  * Read an arbitrary channel setting on the device.
@@ -1251,8 +1323,9 @@ SOAPY_SDR_API char **SoapySDRDevice_listGPIOBanks(const SoapySDRDevice *device,
  * \param device a pointer to a device instance
  * \param bank the name of an available bank
  * \param value an integer representing GPIO bits
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeGPIO(SoapySDRDevice *device, const char *bank, const unsigned value);
+SOAPY_SDR_API int SoapySDRDevice_writeGPIO(SoapySDRDevice *device, const char *bank, const unsigned value);
 
 /*!
  * Write the value of a GPIO bank with modification mask.
@@ -1260,8 +1333,9 @@ SOAPY_SDR_API void SoapySDRDevice_writeGPIO(SoapySDRDevice *device, const char *
  * \param bank the name of an available bank
  * \param value an integer representing GPIO bits
  * \param mask a modification mask where 1 = modify
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeGPIOMasked(SoapySDRDevice *device, const char *bank, const unsigned value, const unsigned mask);
+SOAPY_SDR_API int SoapySDRDevice_writeGPIOMasked(SoapySDRDevice *device, const char *bank, const unsigned value, const unsigned mask);
 
 /*!
  * Readback the value of a GPIO bank.
@@ -1277,8 +1351,9 @@ SOAPY_SDR_API unsigned SoapySDRDevice_readGPIO(const SoapySDRDevice *device, con
  * \param device a pointer to a device instance
  * \param bank the name of an available bank
  * \param dir an integer representing data direction bits
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeGPIODir(SoapySDRDevice *device, const char *bank, const unsigned dir);
+SOAPY_SDR_API int SoapySDRDevice_writeGPIODir(SoapySDRDevice *device, const char *bank, const unsigned dir);
 
 /*!
  * Write the data direction of a GPIO bank with modification mask.
@@ -1287,8 +1362,9 @@ SOAPY_SDR_API void SoapySDRDevice_writeGPIODir(SoapySDRDevice *device, const cha
  * \param bank the name of an available bank
  * \param dir an integer representing data direction bits
  * \param mask a modification mask where 1 = modify
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeGPIODirMasked(SoapySDRDevice *device, const char *bank, const unsigned dir, const unsigned mask);
+SOAPY_SDR_API int SoapySDRDevice_writeGPIODirMasked(SoapySDRDevice *device, const char *bank, const unsigned dir, const unsigned mask);
 
 /*!
  * Read the data direction of a GPIO bank.
@@ -1311,19 +1387,22 @@ SOAPY_SDR_API unsigned SoapySDRDevice_readGPIODir(const SoapySDRDevice *device,
  * \param addr the address of the slave
  * \param data an array of bytes write out
  * \param numBytes the number of bytes to write
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeI2C(SoapySDRDevice *device, const int addr, const char *data, const size_t numBytes);
+SOAPY_SDR_API int SoapySDRDevice_writeI2C(SoapySDRDevice *device, const int addr, const char *data, const size_t numBytes);
 
 /*!
  * Read from an available I2C slave.
  * If the device contains multiple I2C masters,
  * the address bits can encode which master.
+ * Pass the number of bytes to be read in via numBytes;
+ * numBytes will be set to the number of actual bytes read.
  * \param device a pointer to a device instance
  * \param addr the address of the slave
- * \param numBytes the number of bytes to read
+ * \param [inout] numBytes the number of bytes to read
  * \return an array of bytes read from the slave
  */
-SOAPY_SDR_API char *SoapySDRDevice_readI2C(SoapySDRDevice *device, const int addr, const size_t numBytes);
+SOAPY_SDR_API char *SoapySDRDevice_readI2C(SoapySDRDevice *device, const int addr, size_t *numBytes);
 
 /*******************************************************************
  * SPI API
@@ -1365,8 +1444,9 @@ SOAPY_SDR_API char **SoapySDRDevice_listUARTs(const SoapySDRDevice *device, size
  * \param device a pointer to a device instance
  * \param which the name of an available UART
  * \param data a null terminated array of bytes
+ * \return 0 for success or error code on failure
  */
-SOAPY_SDR_API void SoapySDRDevice_writeUART(SoapySDRDevice *device, const char *which, const char *data);
+SOAPY_SDR_API int SoapySDRDevice_writeUART(SoapySDRDevice *device, const char *which, const char *data);
 
 /*!
  * Read bytes from a UART until timeout or newline.
