@@ -4,7 +4,7 @@
 /// Interface definition for Soapy SDR devices.
 ///
 /// \copyright
-/// Copyright (c) 2014-2016 Josh Blum
+/// Copyright (c) 2014-2017 Josh Blum
 /// Copyright (c) 2016-2016 Bastille Networks
 /// SPDX-License-Identifier: BSL-1.0
 ///
@@ -75,9 +75,6 @@ public:
     /*!
      * Unmake or release a device object handle.
      *
-     * \note This call is not thread safe. Implementations calling into unmake
-     * from multiple threads should protect this call with a mutex.
-     *
      * \param device a pointer to a device object
      */
     static void unmake(Device *device);
@@ -161,7 +158,7 @@ public:
      * Query a list of the available stream formats.
      * \param direction the channel direction RX or TX
      * \param channel an available channel on the device
-     * \return a list of allowed format strings
+     * \return a list of allowed format strings. See setupStream() for the format syntax.
      */
     virtual std::vector<std::string> getStreamFormats(const int direction, const size_t channel) const;
 
@@ -190,29 +187,48 @@ public:
      * All stream API calls should be usable with the new stream object
      * after setupStream() is complete, regardless of the activity state.
      *
-     * Format string markup guidelines:
-     *  - C means complex
-     *  - F means floating point
-     *  - S means signed integer
-     *  - U means unsigned integer
-     *  - number float/int size in bytes (complex is 2x this size)
+     * The API allows any number of simultaneous TX and RX streams, but many dual-channel
+     * devices are limited to one stream in each direction, using either one or both channels.
+     * This call will throw an exception if an unsupported combination is requested,
+     * or if a requested channel in this direction is already in use by another stream.
      *
-     * Example format strings:
-     *  - CF32 complex float32 (8 bytes per element)
-     *  - CS16 complex int16 (4 bytes per element)
-     *  - CS12 complex int12 (3 bytes per element)
-     *  - CS4 complex int4 (1 byte per element)
-     *  - S32 int32 (4 bytes per element)
-     *  - U8 uint8 (1 byte per element)
+     * When multiple channels are added to a stream, they are typically expected to have
+     * the same sample rate. See setSampleRate().
      *
-     * Recommended keys to use in the args dictionary:
-     *  - "WIRE" - format of the samples between device and host
+     * \param direction the channel direction (`SOAPY_SDR_RX` or `SOAPY_SDR_TX`)
+     * \param format A string representing the desired buffer format in read/writeStream()
+     * \parblock
      *
-     * \param direction the channel direction RX or TX
-     * \param format the desired buffer format in read/writeStream()
-     * \param channels a list of channels for empty for automatic
-     * \param args stream args or empty for defaults
-     * \return an opaque pointer to a stream handle
+     * The first character selects the number type:
+     *   - "C" means complex
+     *   - "F" means floating point
+     *   - "S" means signed integer
+     *   - "U" means unsigned integer
+     *
+     * The type character is followed by the number of bits per number (complex is 2x this size per sample)
+     *
+     *  Example format strings:
+     *   - "CF32" -  complex float32 (8 bytes per element)
+     *   - "CS16" -  complex int16 (4 bytes per element)
+     *   - "CS12" -  complex int12 (3 bytes per element)
+     *   - "CS4" -  complex int4 (1 byte per element)
+     *   - "S32" -  int32 (4 bytes per element)
+     *   - "U8" -  uint8 (1 byte per element)
+     *
+     * \endparblock
+     * \param channels a list of channels or empty for automatic.
+     * \param args stream args or empty for defaults.
+     * \parblock
+     *
+     *   Recommended keys to use in the args dictionary:
+     *    - "WIRE" - format of the samples between device and host
+     * \endparblock
+     * \return an opaque pointer to a stream handle.
+     * \parblock
+     *
+     * The returned stream is not required to have internal locking, and may not be used
+     * concurrently from multiple threads.
+     * \endparblock
      */
     virtual Stream *setupStream(
         const int direction,
@@ -580,6 +596,30 @@ public:
      */
     virtual std::complex<double> getIQBalance(const int direction, const size_t channel) const;
 
+    /*!
+     * Does the device support frontend frequency correction?
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \return true if frequency corrections are supported
+     */
+    virtual bool hasFrequencyCorrection(const int direction, const size_t channel) const;
+
+    /*!
+     * Fine tune the frontend frequency correction.
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \param value the correction in PPM
+     */
+    virtual void setFrequencyCorrection(const int direction, const size_t channel, const double value);
+
+    /*!
+     * Get the frontend frequency correction value.
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \return the correction value in PPM
+     */
+    virtual double getFrequencyCorrection(const int direction, const size_t channel) const;
+
     /*******************************************************************
      * Gain API
      ******************************************************************/
@@ -795,12 +835,21 @@ public:
 
     /*!
      * Get the range of possible baseband sample rates.
+     * \deprecated replaced by getSampleRateRange()
      * \param direction the channel direction RX or TX
      * \param channel an available channel on the device
      * \return a list of possible rates in samples per second
      */
     virtual std::vector<double> listSampleRates(const int direction, const size_t channel) const;
 
+    /*!
+     * Get the range of possible baseband sample rates.
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \return a list of sample rate ranges in samples per second
+     */
+    virtual RangeList getSampleRateRange(const int direction, const size_t channel) const;
+
     /*******************************************************************
      * Bandwidth API
      ******************************************************************/
@@ -1037,6 +1086,25 @@ public:
      */
     virtual unsigned readRegister(const unsigned addr) const;
 
+    /*!
+     * Write a memory block on the device given the interface name.
+     * This can represent a memory block on a soft CPU, FPGA, IC;
+     * the interpretation is up the implementation to decide.
+     * \param name the name of a available memory block interface
+     * \param addr the memory block start address
+     * \param value the memory block content
+     */
+    virtual void writeRegisters(const std::string &name, const unsigned addr, const std::vector<unsigned> &value);
+
+    /*!
+     * Read a memory block on the device given the interface name.
+     * \param name the name of a available memory block interface
+     * \param addr the memory block start address
+     * \param length number of words to be read from memory block
+     * \return the memory block content
+     */
+    virtual std::vector<unsigned> readRegisters(const std::string &name, const unsigned addr, const size_t length) const;
+
     /*******************************************************************
      * Settings API
      ******************************************************************/
