@@ -1,4 +1,4 @@
-// Copyright (c) 2014-2016 Josh Blum
+// Copyright (c) 2014-2017 Josh Blum
 // Copyright (c) 2016-2016 Bastille Networks
 // SPDX-License-Identifier: BSL-1.0
 
@@ -51,7 +51,7 @@ const char *SoapySDRDevice_lastError(void)
 
 static const bool SoapySDRBoolErr = bool(-1);
 
-static const SoapySDRRange SoapySDRRangeNAN = {NAN, NAN};
+static const SoapySDRRange SoapySDRRangeNAN = {NAN, NAN, 0.0};
 
 static SoapySDRArgInfo SoapySDRArgInfoNull(void)
 {
@@ -164,11 +164,11 @@ int SoapySDRDevice_setupStream(SoapySDRDevice *device, SoapySDRStream **stream,
     //__SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_closeStream(SoapySDRDevice *device, SoapySDRStream *stream)
+int SoapySDRDevice_closeStream(SoapySDRDevice *device, SoapySDRStream *stream)
 {
     __SOAPY_SDR_C_TRY
-    return device->closeStream(reinterpret_cast<SoapySDR::Stream *>(stream));
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->closeStream(reinterpret_cast<SoapySDR::Stream *>(stream));
+    __SOAPY_SDR_C_CATCH
 }
 
 size_t SoapySDRDevice_getStreamMTU(const SoapySDRDevice *device, SoapySDRStream *stream)
@@ -345,13 +345,13 @@ int SoapySDRDevice_setDCOffset(SoapySDRDevice *device, const int direction, cons
     __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_getDCOffset(const SoapySDRDevice *device, const int direction, const size_t channel, double *offsetI, double *offsetQ)
+int SoapySDRDevice_getDCOffset(const SoapySDRDevice *device, const int direction, const size_t channel, double *offsetI, double *offsetQ)
 {
     __SOAPY_SDR_C_TRY
     std::complex<double> ret = device->getDCOffset(direction, channel);
     *offsetI = ret.real();
     *offsetQ = ret.imag();
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    __SOAPY_SDR_C_CATCH
 }
 
 bool SoapySDRDevice_hasIQBalance(const SoapySDRDevice *device, const int direction, const size_t channel)
@@ -368,13 +368,34 @@ int SoapySDRDevice_setIQBalance(SoapySDRDevice *device, const int direction, con
     __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_getIQBalance(const SoapySDRDevice *device, const int direction, const size_t channel, double *balanceI, double *balanceQ)
+int SoapySDRDevice_getIQBalance(const SoapySDRDevice *device, const int direction, const size_t channel, double *balanceI, double *balanceQ)
 {
     __SOAPY_SDR_C_TRY
     std::complex<double> ret = device->getIQBalance(direction, channel);
     *balanceI = ret.real();
     *balanceQ = ret.imag();
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    __SOAPY_SDR_C_CATCH
+}
+
+bool SoapySDRDevice_hasFrequencyCorrection(const SoapySDRDevice *device, const int direction, const size_t channel)
+{
+    __SOAPY_SDR_C_TRY
+    return device->hasFrequencyCorrection(direction, channel);
+    __SOAPY_SDR_C_CATCH_RET(SoapySDRBoolErr);
+}
+
+int SoapySDRDevice_setFrequencyCorrection(SoapySDRDevice *device, const int direction, const size_t channel, const double value)
+{
+    __SOAPY_SDR_C_TRY
+    device->setFrequencyCorrection(direction, channel, value);
+    __SOAPY_SDR_C_CATCH
+}
+
+double SoapySDRDevice_getFrequencyCorrection(const SoapySDRDevice *device, const int direction, const size_t channel)
+{
+    __SOAPY_SDR_C_TRY
+    return device->getFrequencyCorrection(direction, channel);
+    __SOAPY_SDR_C_CATCH_RET(NAN);
 }
 
 /*******************************************************************
@@ -539,6 +560,14 @@ double *SoapySDRDevice_listSampleRates(const SoapySDRDevice *device, const int d
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
+SoapySDRRange *SoapySDRDevice_getSampleRateRange(const SoapySDRDevice *device, const int direction, const size_t channel, size_t *length)
+{
+    *length = 0;
+    __SOAPY_SDR_C_TRY
+    return toRangeList(device->getSampleRateRange(direction, channel), length);
+    __SOAPY_SDR_C_CATCH_RET(nullptr);
+}
+
 /*******************************************************************
  * Bandwidth API
  ******************************************************************/
@@ -659,18 +688,18 @@ long long SoapySDRDevice_getHardwareTime(const SoapySDRDevice *device, const cha
     __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_setHardwareTime(SoapySDRDevice *device, const long long timeNs, const char *what)
+int SoapySDRDevice_setHardwareTime(SoapySDRDevice *device, const long long timeNs, const char *what)
 {
     __SOAPY_SDR_C_TRY
     device->setHardwareTime(timeNs, what);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_setCommandTime(SoapySDRDevice *device, const long long timeNs, const char *what)
+int SoapySDRDevice_setCommandTime(SoapySDRDevice *device, const long long timeNs, const char *what)
 {
     __SOAPY_SDR_C_TRY
     device->setCommandTime(timeNs, what);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    __SOAPY_SDR_C_CATCH
 }
 
 /*******************************************************************
@@ -731,32 +760,35 @@ char **SoapySDRDevice_listRegisterInterfaces(const SoapySDRDevice *device, size_
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_writeNamedRegister(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned value)
+int SoapySDRDevice_writeRegister(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned value)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeRegister(name, addr, value);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeRegister(name, addr, value);
+    __SOAPY_SDR_C_CATCH
 }
 
-unsigned SoapySDRDevice_readNamedRegister(const SoapySDRDevice *device, const char *name, const unsigned addr)
+unsigned SoapySDRDevice_readRegister(const SoapySDRDevice *device, const char *name, const unsigned addr)
 {
     __SOAPY_SDR_C_TRY
     return device->readRegister(name, addr);
     __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_writeRegister(SoapySDRDevice *device, const unsigned addr, const unsigned value)
+int SoapySDRDevice_writeRegisters(SoapySDRDevice *device, const char *name, const unsigned addr, const unsigned *value, const size_t length)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeRegister(addr, value);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeRegisters(name, addr, toNumericVector(value, length));
+    __SOAPY_SDR_C_CATCH
 }
 
-unsigned SoapySDRDevice_readRegister(const SoapySDRDevice *device, const unsigned addr)
+unsigned *SoapySDRDevice_readRegisters(const SoapySDRDevice *device, const char *name, const unsigned addr, size_t *length)
 {
+    const size_t inputLen = *length;
+    *length = 0; //clear in case of error
+
     __SOAPY_SDR_C_TRY
-    return device->readRegister(addr);
-    __SOAPY_SDR_C_CATCH
+    return toNumericList(device->readRegisters(name, addr, inputLen), length);
+    __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
 /*******************************************************************
@@ -770,11 +802,11 @@ SoapySDRArgInfo *SoapySDRDevice_getSettingInfo(const SoapySDRDevice *device, siz
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_writeSetting(SoapySDRDevice *device, const char *key, const char *value)
+int SoapySDRDevice_writeSetting(SoapySDRDevice *device, const char *key, const char *value)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeSetting(key, value);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeSetting(key, value);
+    __SOAPY_SDR_C_CATCH
 }
 
 char *SoapySDRDevice_readSetting(const SoapySDRDevice *device, const char *key)
@@ -792,11 +824,11 @@ SoapySDRArgInfo *SoapySDRDevice_getChannelSettingInfo(const SoapySDRDevice *devi
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_writeChannelSetting(SoapySDRDevice *device, const int direction, const size_t channel, const char *key, const char *value)
+int SoapySDRDevice_writeChannelSetting(SoapySDRDevice *device, const int direction, const size_t channel, const char *key, const char *value)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeSetting(direction, channel, key, value);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeSetting(direction, channel, key, value);
+    __SOAPY_SDR_C_CATCH
 }
 
 char *SoapySDRDevice_readChannelSetting(const SoapySDRDevice *device, const int direction, const size_t channel, const char *key)
@@ -817,18 +849,18 @@ char **SoapySDRDevice_listGPIOBanks(const SoapySDRDevice *device, size_t *length
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_writeGPIO(SoapySDRDevice *device, const char *bank, const unsigned value)
+int SoapySDRDevice_writeGPIO(SoapySDRDevice *device, const char *bank, const unsigned value)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeGPIO(bank, value);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeGPIO(bank, value);
+    __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_writeGPIOMasked(SoapySDRDevice *device, const char *bank, const unsigned value, const unsigned mask)
+int SoapySDRDevice_writeGPIOMasked(SoapySDRDevice *device, const char *bank, const unsigned value, const unsigned mask)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeGPIO(bank, value, mask);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeGPIO(bank, value, mask);
+    __SOAPY_SDR_C_CATCH
 }
 
 unsigned SoapySDRDevice_readGPIO(const SoapySDRDevice *device, const char *bank)
@@ -838,18 +870,18 @@ unsigned SoapySDRDevice_readGPIO(const SoapySDRDevice *device, const char *bank)
     __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_writeGPIODir(SoapySDRDevice *device, const char *bank, const unsigned dir)
+int SoapySDRDevice_writeGPIODir(SoapySDRDevice *device, const char *bank, const unsigned dir)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeGPIODir(bank, dir);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeGPIODir(bank, dir);
+    __SOAPY_SDR_C_CATCH
 }
 
-void SoapySDRDevice_writeGPIODirMasked(SoapySDRDevice *device, const char *bank, const unsigned dir, const unsigned mask)
+int SoapySDRDevice_writeGPIODirMasked(SoapySDRDevice *device, const char *bank, const unsigned dir, const unsigned mask)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeGPIODir(bank, dir, mask);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeGPIODir(bank, dir, mask);
+    __SOAPY_SDR_C_CATCH
 }
 
 unsigned SoapySDRDevice_readGPIODir(const SoapySDRDevice *device, const char *bank)
@@ -862,19 +894,23 @@ unsigned SoapySDRDevice_readGPIODir(const SoapySDRDevice *device, const char *ba
 /*******************************************************************
  * I2C API
  ******************************************************************/
-void SoapySDRDevice_writeI2C(SoapySDRDevice *device, const int addr, const char *data, const size_t numBytes)
+int SoapySDRDevice_writeI2C(SoapySDRDevice *device, const int addr, const char *data, const size_t numBytes)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeI2C(addr, std::string(data, numBytes));
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeI2C(addr, std::string(data, numBytes));
+    __SOAPY_SDR_C_CATCH
 }
 
-char *SoapySDRDevice_readI2C(SoapySDRDevice *device, const int addr, const size_t numBytes)
+char *SoapySDRDevice_readI2C(SoapySDRDevice *device, const int addr, size_t *numBytes)
 {
+    const size_t inputNumBytes = *numBytes;
+    *numBytes = 0; //clear in case of error
+
     __SOAPY_SDR_C_TRY
-    const std::string bytes = device->readI2C(addr, numBytes).c_str();
+    const std::string bytes = device->readI2C(addr, inputNumBytes).c_str();
     char *buff = (char *)std::malloc(bytes.size());
     std::copy(bytes.begin(), bytes.end(), buff);
+    *numBytes = bytes.size();
     return buff;
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
@@ -900,11 +936,11 @@ char **SoapySDRDevice_listUARTs(const SoapySDRDevice *device, size_t *length)
     __SOAPY_SDR_C_CATCH_RET(nullptr);
 }
 
-void SoapySDRDevice_writeUART(SoapySDRDevice *device, const char *which, const char *data)
+int SoapySDRDevice_writeUART(SoapySDRDevice *device, const char *which, const char *data)
 {
     __SOAPY_SDR_C_TRY
-    return device->writeUART(which, data);
-    __SOAPY_SDR_C_CATCH_RET(SoapySDRVoidRet);
+    device->writeUART(which, data);
+    __SOAPY_SDR_C_CATCH
 }
 
 char *SoapySDRDevice_readUART(const SoapySDRDevice *device, const char *which, const long timeoutUs)
